@@ -1,3 +1,4 @@
+from contextlib import contextmanager
 from enum import StrEnum
 from typing import Annotated, Literal
 
@@ -6,6 +7,7 @@ from pydantic import (
     ConfigDict,
     Field,
     PositiveInt,
+    ValidationError,
     field_validator,
     model_validator,
 )
@@ -119,7 +121,7 @@ Faculty is a faculty name.
 """
 
 
-class _StrictBaseModel(BaseModel):
+class StrictBaseModel(BaseModel):
     """
     Base class for all models which need strict validation.
 
@@ -127,13 +129,41 @@ class _StrictBaseModel(BaseModel):
     - model_config: Configuration for the model
     """
 
-    model_config = ConfigDict(extra="forbid", strict=True)
+    model_config = ConfigDict(extra="forbid", strict=True, validate_assignment=True)
     """
-    Configuration for the model which forbids extra fields and is strict (@private)
+    Configuration for the model which forbids extra fields, is strict, and validates on assignment (@private)
     """
 
+    @contextmanager
+    def edit_mode(self):
+        """
+        Context manager for making multiple changes with automatic rollback on validation failure.
 
-class TimeBlock(_StrictBaseModel):
+        **Usage:**
+        ```python
+        with config.edit_mode() as editable_config:
+            editable_config.some_field = "new_value"
+            editable_config.another_field.append("item")
+        # If validation fails, changes are automatically rolled back
+        ```
+
+        **Raises:**
+        - ValueError: If any configuration validation fails (with automatic rollback)
+        """
+        # Create a working copy for editing
+        working_copy = self.model_copy(deep=True)
+        yield working_copy
+        # Validate the working copy by creating a new instance
+        try:
+            validated_copy = self.__class__(**working_copy.model_dump())
+            # If validation passes, update the original object
+            self.__dict__.update(validated_copy.__dict__)
+        except ValidationError as e:
+            # Validation failed, rollback is automatic (working_copy is discarded)
+            raise e
+
+
+class TimeBlock(StrictBaseModel):
     """
     Represents a time block within a day.
     """
@@ -153,24 +183,25 @@ class TimeBlock(_StrictBaseModel):
     End time of the time block
     """
 
-    @field_validator("end")
+    @field_validator("start", "end")
     @classmethod
     def _validate_end_after_start(cls, v, info):
         """
         Validate that the end time is after the start time
         """
-        if "start" in info.data:
+        if "start" in info.data and "end" in info.data:
             start_time = info.data["start"]
+            end_time = info.data["end"]
             # Convert time strings to minutes for comparison
             start_minutes = int(start_time.split(":")[0]) * 60 + int(start_time.split(":")[1])
-            end_minutes = int(v.split(":")[0]) * 60 + int(v.split(":")[1])
+            end_minutes = int(end_time.split(":")[0]) * 60 + int(end_time.split(":")[1])
 
             if end_minutes <= start_minutes:
                 raise ValueError("End time must be after start time")
         return v
 
 
-class TimeRange(_StrictBaseModel):
+class TimeRange(StrictBaseModel):
     """
     A time range with start and end times, ensuring start < end.
     """
@@ -212,7 +243,7 @@ class TimeRange(_StrictBaseModel):
         return cls(start=start, end=end)
 
 
-class Meeting(_StrictBaseModel):
+class Meeting(StrictBaseModel):
     """
     Represents a single meeting instance.
     """
@@ -238,7 +269,7 @@ class Meeting(_StrictBaseModel):
     """
 
 
-class ClassPattern(_StrictBaseModel):
+class ClassPattern(StrictBaseModel):
     """
     Represents a class pattern.
     """
@@ -281,7 +312,7 @@ class ClassPattern(_StrictBaseModel):
         return v
 
 
-class TimeSlotConfig(_StrictBaseModel):
+class TimeSlotConfig(StrictBaseModel):
     """
     Represents a time slot configuration.
     """
@@ -316,8 +347,41 @@ class TimeSlotConfig(_StrictBaseModel):
     Minimum time overlap between time slots (default: 45)
     """
 
+    @model_validator(mode="after")
+    def validate(self):
+        """
+        Validate that time slot config is consistent and complete.
+        """
+        errors = []
 
-class CourseConfig(_StrictBaseModel):
+        # Check that all days in time_slot_config are valid
+        valid_days = {"MON", "TUE", "WED", "THU", "FRI"}
+        for day in self.times:
+            if day not in valid_days:
+                errors.append(f"Invalid day '{day}' in time slot configuration")
+
+        # Check that there are time blocks for each day
+        for day in valid_days:
+            if day not in self.times or not self.times[day]:
+                errors.append(f"No time blocks defined for {day}")
+
+        # Check that class patterns are reasonable
+        if not self.classes:
+            errors.append("At least one class pattern must be defined")
+
+        # Check for disabled patterns
+        disabled_patterns = [p for p in self.classes if p.disabled]
+        if len(disabled_patterns) == len(self.classes):
+            errors.append("All class patterns are disabled")
+
+        if errors:
+            error_message = "Time slot configuration errors:\n" + "\n".join(f"  - {error}" for error in errors)
+            raise ValueError(error_message)
+
+        return self
+
+
+class CourseConfig(StrictBaseModel):
     """
     Represents a course configuration.
     """
@@ -352,14 +416,8 @@ class CourseConfig(_StrictBaseModel):
     List of faculty names
     """
 
-    @model_validator(mode="after")
-    def _validate_references(self):
-        """Validate that all references exist in the parent SchedulerConfig.
-        This validator will be called by the parent SchedulerConfig."""
-        return self
 
-
-class FacultyConfig(_StrictBaseModel):
+class FacultyConfig(StrictBaseModel):
     """
     Represents a faculty configuration.
     """
@@ -441,9 +499,9 @@ class FacultyConfig(_StrictBaseModel):
         return v
 
     @model_validator(mode="after")
-    def _validate_credit_consistency(self):
+    def validate(self):
         """
-        Validate that minimum and maximum credits are consistent
+        Validate the model state.
         """
         if self.minimum_credits > self.maximum_credits:
             raise ValueError(
@@ -453,7 +511,7 @@ class FacultyConfig(_StrictBaseModel):
         return self
 
 
-class SchedulerConfig(_StrictBaseModel):
+class SchedulerConfig(StrictBaseModel):
     """
     Represents a scheduler configuration.
     """
@@ -505,10 +563,19 @@ class SchedulerConfig(_StrictBaseModel):
     """
 
     @model_validator(mode="after")
-    def _validate_all_references(self):
+    def validate(self):
         """
-        Validate that all `Faculty`, `Room`, `Lab`, and `Course` references exist.
-        Validate that all `Faculty`, `Room`, and `Lab` definitions are unique.
+        Validate all cross-references between child models.
+        This method can be called manually or is used by Pydantic validators.
+
+        **Usage:**
+        ```python
+        config.courses[0].room = ["NewRoom"]
+        config.validate()  # Validates all cross-references
+        ```
+
+        **Raises:**
+        - ValueError: If any cross-reference validation fails
         """
         # Validate uniqueness first
         self._validate_uniqueness()
@@ -652,7 +719,7 @@ class OptimizerFlags(StrEnum):
     """
 
 
-class CombinedConfig(_StrictBaseModel):
+class CombinedConfig(StrictBaseModel):
     """
     Represents a combined configuration.
     """
@@ -693,7 +760,13 @@ class CombinedConfig(_StrictBaseModel):
     time_slot_config: TimeSlotConfig = Field(
         description="Time slot configuration",
         example=TimeSlotConfig(
-            times={"MON": [{"start": "10:00", "spacing": 60, "end": "12:00"}]},
+            times={
+                "MON": [{"start": "10:00", "spacing": 60, "end": "12:00"}],
+                "TUE": [{"start": "10:00", "spacing": 60, "end": "12:00"}],
+                "WED": [{"start": "10:00", "spacing": 60, "end": "12:00"}],
+                "THU": [{"start": "10:00", "spacing": 60, "end": "12:00"}],
+                "FRI": [{"start": "10:00", "spacing": 60, "end": "12:00"}],
+            },
             classes=[{"credits": 3, "meetings": [{"day": "MON", "duration": 150, "lab": False}]}],
         ),
     )
@@ -732,36 +805,3 @@ class CombinedConfig(_StrictBaseModel):
         if isinstance(v, list):
             return [OptimizerFlags(flag) if isinstance(flag, str) else flag for flag in v]
         return v
-
-    @model_validator(mode="after")
-    def _validate_time_slot_config_consistency(self):
-        """
-        Validate that time slot config is consistent with scheduler config
-        """
-        errors = []
-
-        # Check that all days in time_slot_config are valid
-        valid_days = {"MON", "TUE", "WED", "THU", "FRI"}
-        for day in self.time_slot_config.times:
-            if day not in valid_days:
-                errors.append(f"Invalid day '{day}' in time slot configuration")
-
-        # Check that there are time blocks for each day
-        for day in valid_days:
-            if day not in self.time_slot_config.times or not self.time_slot_config.times[day]:
-                errors.append(f"No time blocks defined for {day}")
-
-        # Check that class patterns are reasonable
-        if not self.time_slot_config.classes:
-            errors.append("At least one class pattern must be defined")
-
-        # Check for disabled patterns
-        disabled_patterns = [p for p in self.time_slot_config.classes if p.disabled]
-        if len(disabled_patterns) == len(self.time_slot_config.classes):
-            errors.append("All class patterns are disabled")
-
-        if errors:
-            error_message = "Time slot configuration errors:\n" + "\n".join(f"  - {error}" for error in errors)
-            raise ValueError(error_message)
-
-        return self
